--
-- Keystone Communication Library by Strucker
--
-- Shares keystone and character information across Guild and Group addon channels, and among friends list members.

-- Required Libraries:
--   * LibStub
--   * LibRealmInfo
--   * ChatThrottlePlusLib
-- 
-- Required libraries must be loaded BEFORE this one.
-- 
-- To receive keystone entries as they are received by the library, add a listener function:
--
--    KCLib:AddKeystoneListener(function(keystoneEntry,channel,sender) 
--       --Your Logic Here
--    end)
-- 
--    Listeners are passed:
--    1. A Keystone Entry (See Sample Below)
--    2. The channel the keystoneEntry was received through. (WHISPER, PARTY, RAID, INSTANCE_CHAT, GUILD, BNET, PLAYER)
--    3. The character who provided the keystoneEntry information, this will not always be the same as the character the keystoneEntry is about.
--
-- To get the player's most recent keystone entry (Player keystone entries will come through the listener function as well):
--
--    KCLib:GetPlayerKeystone()
--
-- To add keystone entries for the player's alts:
--
--    KCLib:AddAltKeystone(keystoneEntry)
--
-- Sample Keystone Entry:
-- {
--    ["guid"] = "4-0A810590",
--			-- Sourced from UnitGuid. Is guaranteed to remain constant from character creation to character deletion, faction change, or server transfer.
--    ["name"] = "Struckor-Kilrogg",
--    ["realm"] = "Kilrogg",
--    ["level"] = 110,
--    ["faction"] = 1,
--			-- 1 for alliance, 0 for horde.
--    ["class"] = "DEATHKNIGHT",
--			-- Not localized
--    ["guildName"] = "Transcendence",
--    ["ilvl"] = 999,
--    ["ilvlEquipped"] = 998,
--    ["keystoneWeekIndex"] = 100,
--			-- Used to identify which weekly reset the keystone entry was created during.
--    ["keystoneDungeonId"] = 197,
--			-- Use 'C_ChallengeMode.GetMapInfo(keystoneDungeonId)' to get the dungeon name.
--    ["keystoneLevel"] = 19,
--    ["bestKeystoneDungeonId"] = 207,
--			-- Use 'C_ChallengeMode.GetMapInfo(keystoneDungeonId)' to get the dungeon name.
--    ["bestKeystoneLevel"] = 18,
--    ["altOf"] = "Shastrucker-Kilrogg",
--			-- Can be the same as 'name'.
--    ["lastSeen"] = 99999999999 
--			-- The time in seconds that the keystone entry was last received.
-- }
--
-- Embeddable library. Can run as a standalone addon also, but, really, just embed it!
--

local LIBRARY_NAME = "Keystone Communication Library";
local VERSION = 1.63;
local MSG_PREFIX = "KCLib";
local DELAY_LENGTH = 2;
local WEEK_SECONDS = 604800;
local RESET_WEDNESDAY = 1500447600;
local RESET_TUESDAY = 1500390000;
local KEYSTONE_ITEM_ID = 158923;
local ALT_KEYSTONES = {};
local LISTENERS = {};
local PLAYER_NAME,PLAYER_REALM,PLAYER_GUID;
local LAUNCH_TIME = LibStub('LibRealmInfo'):GetCurrentRegion() == "EU" and RESET_WEDNESDAY or RESET_TUESDAY;
local playerKeystone;
local loaded = false;
local delayedRun = 0;
local activeMythicPlusMap, deathCount, deathTimePenalty = 0, 0, 0;

local keystoneUpgradeTimePoints = {
	[244] = {1800,1440,1080},
	[245] = {2160,1728,1296},
	[246] = {2160,1728,1296},
	[247] = {2340,1908,1404},
	[248] = {2340,1908,1404},
	[249] = {2340,1908,1404},
	[250] = {2160,1728,1296},
	[251] = {1980,1584,1188},
	[252] = {2340,1908,1404},
	[353] = {2160,1728,1296}
}

-- Track which chat channels we are fully synced with.
local guildSynced, groupSynced, friendsSynced = false, false, false;

-- Ensure LibStub is setup.
if not LibStub then
	error(LIBRARY_NAME.. " requires LibStub.");
	return ;
end

-- Ensure ChatThrottlePlusLib is setup.
if not ChatThrottlePlusLib then
	error(LIBRARY_NAME.. " requires ChatThrottlePlusLib.");
	return ;
end

-- Ensure another instance of the library isn't already setup.
local KCLib = LibStub:NewLibrary(LIBRARY_NAME,VERSION)
if not KCLib then 
	return;
end

_G.KCLib = KCLib;

------------------------------------------------------------------------------------------------
-- Add a listener function to be called whenever a new keystone record is received.
------------------------------------------------------------------------------------------------
function KCLib:AddKeystoneListener(listener)
	tinsert(LISTENERS,listener);
end

------------------------------------------------------------------------------------------------
-- Add keystone entries for the player's alternate characters.
--    Alt keystone entries will be shared across chat channels along with the player's keystone.
------------------------------------------------------------------------------------------------
function KCLib:AddAltKeystone(keystone)
	keystone.altOf = PLAYER_NAME;
	keystone.seasonScore = keystone.seasonScore or 0;
	tinsert(ALT_KEYSTONES,keystone);
end

------------------------------------------------------------------------------------------------
-- Sets the player keystone entry. This will be overriden with the next keystone update.
------------------------------------------------------------------------------------------------
function KCLib:SetPlayerKeystone(keystone)
	keystone.keystoneDungeonId = keystone.keystoneDungeonId or 0;
	keystone.keystoneLevel = keystone.keystoneLevel or 0;
	keystone.bestKeystoneDungeonId = keystone.bestKeystoneDungeonId or 0;
	keystone.bestKeystoneLevel = keystone.bestKeystoneLevel or 0;
	playerKeystone = keystone
end

------------------------------------------------------------------------------------------------
-- Get the player's current group chat channel.
------------------------------------------------------------------------------------------------
function KCLib:GetGroupChatChannel()
	if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
		return "INSTANCE_CHAT";
	elseif UnitInRaid("player") then
		return "RAID";
	elseif IsInGroup() then
		return "PARTY";
	end
	return nil;
end

-------------------------------------------------------------------------------------
-- Get the time in seconds, for the current region, when the mythic affix schedule was launch.
-------------------------------------------------------------------------------------
function KCLib:GetLaunchTime()
	return LAUNCH_TIME;
end

-------------------------------------------------------------------------------------
-- How many weeks into the mythic+ affix schedule are we?
-------------------------------------------------------------------------------------
function KCLib:GetWeeklyIndex() 
	return math.floor((GetServerTime() - LAUNCH_TIME)/ WEEK_SECONDS);
end

-------------------------------------------------------------------------------------
-- Get the player's full name (name-realm)
-------------------------------------------------------------------------------------
function KCLib:GetFullPlayerName()
	return PLAYER_NAME;
end

-------------------------------------------------------------------------------------
-- Get the player's home realm
-------------------------------------------------------------------------------------
function KCLib:GetPlayerRealm()
	return PLAYER_REALM;
end

-------------------------------------------------------------------------------------
-- Calculate the score Bonus/Penalty.
-- 1 point per 3 minutes above/below the time limit, to a maximum of +/- 5 points.
-- The cap of a 5 point bonus/penalty ensure a higher level run's score is always 
--     greather then or equal to the score of all lower level runs' scores.
-------------------------------------------------------------------------------------
function KCLib:GetScoreTimeAdjustment(timeLimit, completionTime)
	return math.floor(math.max(math.min( (timeLimit - completionTime) / 180.0,5),-5)*10)/10;
end


-------------------------------------------------------------------------------------
-- Calculate the score of a particuler run.
-- 10 points per keystone level
-- +/- up to 5 points based on completion time.
-------------------------------------------------------------------------------------
local function GetRunScore(dungeonId, level, completionTime)
	if not level or level <= 0 then
		return 0,0,0;
	end
	
	local _, _, timeLimit = C_ChallengeMode.GetMapUIInfo(dungeonId);
	local timeDelta = KCLib:GetScoreTimeAdjustment(timeLimit, completionTime);
	local levelScore = level * 10;
	return levelScore + timeDelta, levelScore, timeDelta;
end


-------------------------------------------------------------------------------------
-- Compile the keystones local run history, C_MythicPlus.GetSeasonBestForMap, and C_MythicPlus.GetWeeklyBestForMap.
-- returns the weekly and season best run information.
-------------------------------------------------------------------------------------
local function GetRunStatsForMap(keystone, dungeonId)
	local _, _, timeLimit = C_ChallengeMode.GetMapUIInfo(dungeonId);
	local weekBestRunInfo = {}
	local seasonBestRunInfo = {}
	local timePoints = keystoneUpgradeTimePoints[dungeonId];
	
	local inTimeInfo, overtimeInfo = C_MythicPlus.GetSeasonBestForMap(dungeonId);
	weekBestRunInfo.completionTime, weekBestRunInfo.level, _, _, members = C_MythicPlus.GetWeeklyBestForMap(dungeonId);
	weekBestRunInfo.completionTime = weekBestRunInfo.completionTime or 0;
	weekBestRunInfo.level = weekBestRunInfo.level or 0;
	weekBestRunInfo.score, weekBestRunInfo.levelScore, weekBestRunInfo.timeScore = GetRunScore(dungeonId, weekBestRunInfo.level, weekBestRunInfo.completionTime);
	if weekBestRunInfo.completionTime > timePoints[1] then
		weekBestRunInfo.keystoneUpgradeLevels = 0;
	elseif weekBestRunInfo.completionTime > timePoints[2] then
		weekBestRunInfo.keystoneUpgradeLevels = 1;
	elseif weekBestRunInfo.completionTime > timePoints[3] then
		weekBestRunInfo.keystoneUpgradeLevels = 2;
	else
		weekBestRunInfo.keystoneUpgradeLevels = 3;
	end
	weekBestRunInfo.partyMembers = members;
	
	local info = inTimeInfo or overtimeInfo or nil;
	if info then
		seasonBestRunInfo.level = info.level or 0
		seasonBestRunInfo.completionTime = info.durationSec or 0
		seasonBestRunInfo.score, seasonBestRunInfo.levelScore, seasonBestRunInfo.timeScore = GetRunScore(dungeonId, seasonBestRunInfo.level, seasonBestRunInfo.completionTime);
		if seasonBestRunInfo.completionTime > timePoints[1] then
			seasonBestRunInfo.keystoneUpgradeLevels = 0;
		elseif seasonBestRunInfo.completionTime > timePoints[2] then
			seasonBestRunInfo.keystoneUpgradeLevels = 1;
		elseif seasonBestRunInfo.completionTime > timePoints[3] then
			seasonBestRunInfo.keystoneUpgradeLevels = 2;
		else
			seasonBestRunInfo.keystoneUpgradeLevels = 3;
		end
		seasonBestRunInfo.partyMembers = info.members;
	else
		seasonBestRunInfo.level = 0;
		seasonBestRunInfo.completionTime = 0;
		seasonBestRunInfo.score = 0;
		seasonBestRunInfo.levelScore = 0;
		seasonBestRunInfo.timeScore = 0;
		seasonBestRunInfo.keystoneUpgradeLevels = 0;
	end
	
	if weekBestRunInfo.score > seasonBestRunInfo.score then
		seasonBestRunInfo.level = weekBestRunInfo.level;
		seasonBestRunInfo.completionTime = weekBestRunInfo.completionTime;
		seasonBestRunInfo.score = weekBestRunInfo.score;
		seasonBestRunInfo.levelScore = weekBestRunInfo.levelScore;
		seasonBestRunInfo.timeScore = weekBestRunInfo.timeScore;
		seasonBestRunInfo.keystoneUpgradeLevels = weekBestRunInfo.keystoneUpgradeLevels;
		seasonBestRunInfo.partyMembers = weekBestRunInfo.partyMembers;
	end
	seasonBestRunInfo.count = 0;
	weekBestRunInfo.count = 0;
	for i,runInfo in ipairs(keystone.runHistory) do
		if runInfo.level > 0 and runInfo.dungeonId == dungeonId and runInfo.seasonId == keystone.seasonId then
			local runScore, runLevelScore, runTimeScore = GetRunScore(dungeonId, runInfo.level, runInfo.completionTime);
			seasonBestRunInfo.count = seasonBestRunInfo.count +1;
			if runScore >= seasonBestRunInfo.score then
				seasonBestRunInfo.level = runInfo.level;
				seasonBestRunInfo.completionTime = runInfo.completionTime;
				seasonBestRunInfo.score = runScore;
				seasonBestRunInfo.keystoneUpgradeLevels = runInfo.keystoneUpgradeLevels;
				seasonBestRunInfo.levelScore = runLevelScore;
				seasonBestRunInfo.timeScore = runTimeScore;
				seasonBestRunInfo.partyMembers = runInfo.partyMembers
			end
			
			if runInfo.weekIndex == keystone.keystoneWeekIndex then
				weekBestRunInfo.count = weekBestRunInfo.count + 1;
				if runScore >= weekBestRunInfo.score then
					weekBestRunInfo.level = runInfo.level;
					weekBestRunInfo.completionTime = runInfo.completionTime;
					weekBestRunInfo.score = runScore;
					weekBestRunInfo.keystoneUpgradeLevels = runInfo.keystoneUpgradeLevels;
					weekBestRunInfo.levelScore = runLevelScore;
					weekBestRunInfo.timeScore = runTimeScore;
					weekBestRunInfo.partyMembers = runInfo.partyMembers
				end
			end
		end
	end
	
	return weekBestRunInfo, seasonBestRunInfo;
end

------------------------------------------------------------------------------------------------
-- Build a new keystone record for the player.
--   If this is the first build of the session or the keystone entry has changed,
--   the new keystone entry will be forwarded to listeners.
------------------------------------------------------------------------------------------------
local function RefreshPlayerKeystone()
	local keystone = {};
	local weekIndex = KCLib:GetWeeklyIndex();
	if playerKeystone and playerKeystone.keystoneWeekIndex == weekIndex then
		keystone.bestKeystoneDungeonId = playerKeystone.bestKeystoneDungeonId;
		keystone.bestKeystoneTime = playerKeystone.bestKeystoneTime;
		keystone.bestKeystoneLevel = playerKeystone.bestKeystoneLevel;
	end
	
	keystone.runHistory = playerKeystone and playerKeystone.runHistory or {}
	keystone.altered = false;
	keystone.guid = PLAYER_GUID;
	keystone.name = PLAYER_NAME;
	keystone.altOf = PLAYER_NAME;
	keystone.realm = PLAYER_REALM;
	keystone.race = select(2,UnitRace("player"));
	keystone.raceId = select(3,UnitRace("player"));
	keystone.level = UnitLevel("player");
	keystone.faction = UnitFactionGroup("player") == "Alliance";
	keystone.class = select(2,UnitClass("player"));
	keystone.classId = select(3,UnitClass("player"));
	keystone.guildName = GetGuildInfo("player");
	local ilvl, ilvlEquipped = GetAverageItemLevel();
	keystone.ilvl = math.floor(ilvl) ;
	keystone.ilvlEquipped = math.floor(ilvlEquipped);
	keystone.keystoneWeekIndex = weekIndex;
	keystone.keystoneDungeonId = C_MythicPlus.GetOwnedKeystoneChallengeMapID() or 0;
	keystone.keystoneLevel = C_MythicPlus.GetOwnedKeystoneLevel() or 0;
	keystone.seasonId = C_MythicPlus.GetCurrentSeason()
	keystone.weekRunInfos = {};
	keystone.seasonRunInfos = {};
	keystone.bestKeystoneDungeonId = 0;
	keystone.bestKeystoneTime = 0;
	keystone.bestKeystoneLevel = 0;
	keystone.bestKeystoneScore = 0;
	
	local maps = C_ChallengeMode.GetMapTable();
	local scoreTotal = 0;
	for i, dungeonId in ipairs(maps) do
		keystone.weekRunInfos[dungeonId], keystone.seasonRunInfos[dungeonId] = GetRunStatsForMap(keystone, dungeonId)		
		scoreTotal = scoreTotal + keystone.seasonRunInfos[dungeonId].score;
		if keystone.weekRunInfos[dungeonId].score >= keystone.bestKeystoneScore then
			keystone.bestKeystoneDungeonId = dungeonId;
			keystone.bestKeystoneLevel = keystone.weekRunInfos[dungeonId].level;
			keystone.bestKeystoneTime = keystone.weekRunInfos[dungeonId].completionTime;
			keystone.bestKeystoneScore = keystone.weekRunInfos[dungeonId].score
		end
    end
	keystone.seasonScore = math.floor(scoreTotal);
	
	local unopendChest = C_MythicPlus.IsWeeklyRewardAvailable()
	local currentWeekBestLevel, weeklyRewardLevel, nextDifficultyWeeklyRewardLevel, nextBestLevel = C_MythicPlus.GetWeeklyChestRewardLevel();
	if not unopendChest and currentWeekBestLevel and currentWeekBestLevel > keystone.bestKeystoneLevel then
		keystone.bestKeystoneDungeonId = 0;
		keystone.bestKeystoneLevel = currentWeekBestLevel or 0;
		keystone.bestKeystoneTime = 0;
	end
	
	-- Check if the keystone entry has changed.
	local changed = false;
	if playerKeystone then
		for key,val in pairs(keystone) do
			if keystone[key] ~= playerKeystone[key] then
				changed = true;
				break;
			end
		end
	else
		changed = true;
	end
	keystone.lastSeen = GetServerTime();
	
	-- If the keystone changed, send it to the listeners.
	if changed then
		for _,listener in ipairs(LISTENERS) do
			listener(keystone,"PLAYER",PLAYER_NAME);
		end
	end
	
	playerKeystone = keystone;
	return changed;
end

------------------------------------------------------------------------------------------------
-- Fetch the latest player keystone entry.
--    Use the listener to be informed of updates to the player keystone entry.
------------------------------------------------------------------------------------------------
function KCLib:GetPlayerKeystone()
	if not playerKeystone then
		RefreshPlayerKeystone();
	end
	return playerKeystone;
end

------------------------------------------------------------------------------------------------
-- Share the given keystone over the given chat channel.
-- Max Character Length: 213
-- Never Exceed 249 or we risk crashing the user's game!
------------------------------------------------------------------------------------------------
local function SendKeystone(requestSync,keystone,channel,target)
	if keystone then
		local msgArray = {};
		msgArray[1] = requestSync;
		msgArray[2] = keystone.guid or "";
		msgArray[3] = keystone.name or "";
		msgArray[4] = keystone.level or "";
		msgArray[5] = keystone.faction and "1" or "0";
		msgArray[6] = keystone.classId or "";
		msgArray[7] = keystone.guildName or "";
		msgArray[8] = keystone.ilvl or "";
		msgArray[9] = keystone.ilvlEquipped or "";
		msgArray[10] = keystone.keystoneWeekIndex or "";
		msgArray[11] = keystone.keystoneDungeonId or "";
		msgArray[12] = keystone.keystoneLevel or "";
		msgArray[13] = keystone.bestKeystoneDungeonId or "";
		msgArray[14] = keystone.bestKeystoneLevel or "";
		msgArray[15] = keystone.raceId or "";
		msgArray[16] = "";
		msgArray[17] = keystone.seasonId or "";
		local maps = C_ChallengeMode.GetMapTable();
		if keystone.seasonRunInfos then
			for i,dungeonId in ipairs(maps) do
				local runInfo = keystone.seasonRunInfos[dungeonId];
				if runInfo then
					msgArray[17 + i] = runInfo.level
					msgArray[18 + #maps + i] = runInfo.keystoneUpgradeLevels
					msgArray[18 + (#maps * 2) + i] = math.floor(runInfo.timeScore * 10)
				else
					msgArray[17 + i] = 0
					msgArray[18 + #maps + i] = 0
					msgArray[18 + (#maps * 2) + i] = 0
				end
			end
		end
		msgArray[18 + #maps] = keystone.seasonScore or 0;

		local msg = table.concat(msgArray, ",") 
		if channel == "BNET" then
			ChatThrottlePlusLib:BNSendGameData("BULK",MSG_PREFIX, msg, target);
		else
			ChatThrottlePlusLib:SendAddonMessage("BULK",MSG_PREFIX, msg, channel, target);
		end
	end
end

------------------------------------------------------------------------------------------------
-- Share the player keystone and all alt keystones to the given channel.
------------------------------------------------------------------------------------------------
local function SendAllKeystones(requestSync,channel,target)
	if playerKeystone then
		SendKeystone(requestSync,playerKeystone,channel,target);
	end
	if ALT_KEYSTONES then
		for _,keystone in ipairs(ALT_KEYSTONES) do
			SendKeystone(0,keystone,channel,target);
		end
	end
end

------------------------------------------------------------------------------------------------
-- Build a keystone out of the incoming message and send it to all listeners.
------------------------------------------------------------------------------------------------
local function ProcessKeystoneMessage(msgArray,channel,sender)
	local keystone = {}	;
	keystone.guid = msgArray[2];
	keystone.name = msgArray[3];
	keystone.realm = select(2,strsplit(keystone.name,"-"));
	keystone.level = tonumber(msgArray[4]) or 0;
	keystone.faction = msgArray[5] == "1";
	--Maintain the ability to extract class names from messages sent by older versions of the library.
	if msgArray[6] and tonumber(msgArray[6]) and tonumber(msgArray[6]) > 0 then
		keystone.classId = tonumber(msgArray[6]);
		keystone.class = select(2,GetClassInfo(keystone.classId))
	else
		keystone.classId = 0;
		keystone.class = msgArray[6] or "UNKNOWN"
	end
	keystone.guildName = msgArray[7];
	keystone.ilvl = tonumber(msgArray[8]) or 0;
	keystone.ilvlEquipped = tonumber(msgArray[9]) or 0;
	keystone.keystoneWeekIndex = tonumber(msgArray[10]) or 0;
	keystone.keystoneDungeonId = tonumber(msgArray[11]) or 0;
	keystone.keystoneLevel = tonumber(msgArray[12]) or 0;
	keystone.bestKeystoneDungeonId = tonumber(msgArray[13]) or 0;
	keystone.bestKeystoneLevel = tonumber(msgArray[14]) or 0;
	if msgArray[15] and tonumber(msgArray[15]) and tonumber(msgArray[15]) > 0 then
		keystone.raceId = tonumber(msgArray[15]);
		local raceRecord = C_CreatureInfo.GetRaceInfo(keystone.raceId)
		keystone.race = raceRecord and raceRecord.raceName or "UNKNOWN"
	else
		keystone.raceId = 0;
		keystone.race = msgArray[15] or "UNKNOWN"
	end
	keystone.seasonId =	tonumber(msgArray[17]) or 0;
	if #msgArray > 17 then
		keystone.seasonRunInfos = {};
		local maps = C_ChallengeMode.GetMapTable();
		for i,dungeonId in ipairs(maps) do
			local runInfo = {}
			runInfo.level = tonumber(msgArray[17 + i]) or 0
			runInfo.keystoneUpgradeLevels = tonumber(msgArray[18 + #maps + i]) or 0
			runInfo.timeScore = (tonumber(msgArray[18 + (#maps * 2) + i]) or 0) / 10
			runInfo.levelScore = runInfo.level * 10
			runInfo.score = runInfo.levelScore + runInfo.timeScore
			keystone.seasonRunInfos[dungeonId] = runInfo
		end
		keystone.seasonScore = tonumber(msgArray[18 + #maps]) or 0;
	end
	keystone.seasonScore = keystone.seasonScore or 0;
	keystone.altOf = sender;
	keystone.lastSeen = GetServerTime();
	-- Insure the keystone is not outdated.
	if keystone.keystoneWeekIndex == KCLib:GetWeeklyIndex() then
		-- Send to all listeners.
		for _,listener in ipairs(LISTENERS) do
			listener(keystone,channel,sender);
		end
	end
end

local function BroadcastPlayerKeys()
	local isNew = RefreshPlayerKeystone();
	-- Send keystone changes to guild and resync if needed.
	if GetGuildInfo("player") then
		if not guildSynced then
			SendAllKeystones(1,"GUILD");
		elseif isNew then
			SendKeystone(0, playerKeystone, "GUILD");
		end
		guildSynced = true;
	end
	
	-- Send keystone changes to group and resync if needed
	local groupChatChannel = KCLib:GetGroupChatChannel();
	if groupChatChannel then
		if not groupSynced then
			SendAllKeystones(1,groupChatChannel);
		elseif isNew then
			SendKeystone(0, playerKeystone, groupChatChannel);
		end
		groupSynced = true;
	end
	
	-- Send keystone changes to friends and resync if needed
	if isNew or not friendsSynced then
		for i = 1, select(2, GetNumFriends()) do
			local fullName = GetFriendInfo(i);
			if not string.match(fullName, "-") then
				fullName = fullName .. "-" .. PLAYER_REALM;
			end
			
			if friendsSynced then
				SendKeystone(0, playerKeystone, "WHISPER",fullName);
			else
				SendAllKeystones(1,"WHISPER",fullName);
			end
		end
		
		if BNConnected() then
			for i = 1, BNGetNumFriends() do
				local characterName, bnetIDGameAccount, client = select(5,BNGetFriendInfo(i))
				if bnetIDGameAccount and characterName and client == "WoW" then
					if friendsSynced then
						SendKeystone(0, playerKeystone, "BNET", bnetIDGameAccount);
					else
						SendAllKeystones(1, "BNET", bnetIDGameAccount);
					end
				end
			end
		end
		friendsSynced = true;
	end
end

local EventCallables = {}

-- On login, register addon prefix and check player name/realm.
EventCallables["PLAYER_LOGIN"] = function()
	local name,realm = UnitFullName("player");
	PLAYER_NAME = name .."-" .. realm;
	PLAYER_REALM = realm;
	
	PLAYER_GUID = string.sub(UnitGUID("player"),8);
	
	C_ChatInfo.RegisterAddonMessagePrefix(MSG_PREFIX);	
end

-- When joining a guild mark the guild chat unsynced.
EventCallables["PLAYER_GUILD_UPDATE"] = function()
	guildSynced = false;
	C_MythicPlus.RequestRewards();
	C_MythicPlus.RequestMapInfo();
end

-- When joining a group mark the group chat unsynced.
EventCallables["GROUP_JOINED"] = function()
	groupSynced = false;
	C_MythicPlus.RequestRewards();
	C_MythicPlus.RequestMapInfo();
end

-- When player first enters the world, do a full sync.
EventCallables["PLAYER_ENTERING_WORLD"] = function()
	groupSynced = false;
	guildSynced = false;
	friendsSynced = false;
	C_MythicPlus.RequestRewards();
	C_MythicPlus.RequestMapInfo();
end

-- Schedule a check of the player keystone.
EventCallables["BAG_UPDATE"] = function()
	delayedRun = time() + DELAY_LENGTH;
end

-- Possable run completion, or fresh load, schedule new player keystone broadcast
EventCallables["CHALLENGE_MODE_MAPS_UPDATE"] = function()
	delayedRun = time() + DELAY_LENGTH;
end

-- Process incoming addon messages and extract keystones.
EventCallables["CHAT_MSG_ADDON"] = function(prefix, msg, channel, sender)
	if prefix == MSG_PREFIX and sender ~= PLAYER_NAME then
		local msgParts =  {strsplit(",",msg)};
		if msgParts[1] == "1" then
			SendAllKeystones(0,channel, sender);
		end
		ProcessKeystoneMessage(msgParts,channel,sender);
	end
end

-- Process incoming addon messages and extract keystones.
EventCallables["BN_CHAT_MSG_ADDON"] = function(prefix, msg, channel, sender)
	if prefix == MSG_PREFIX then
		local msgParts =  {strsplit(",",msg)};
		if msgParts[1] == "1" then
			SendAllKeystones(0,"BNET", sender);
		end
		local characterName, _, realm = select(2,BNGetGameAccountInfo(sender));
		ProcessKeystoneMessage(msgParts,"BNET",characterName .. "-" .. realm);
	end
end

EventCallables["CHALLENGE_MODE_RESET"] = function()
	C_MythicPlus.RequestRewards();
	C_MythicPlus.RequestMapInfo();
end

EventCallables["CHALLENGE_MODE_START"] = EventCallables["CHALLENGE_MODE_RESET"]
EventCallables["CHALLENGE_MODE_MEMBER_INFO_UPDATED"] = EventCallables["CHALLENGE_MODE_RESET"]

EventCallables["CHALLENGE_MODE_COMPLETED"] = function()
	local dungeonId, level, completionTime, onTime, keystoneUpgradeLevels = C_ChallengeMode.GetCompletionInfo();
	local enemyDamagerModifier, enemyHealthModifier = C_ChallengeMode.GetPowerLevelDamageHealthMod(level);
	local partyMembers = {}
	
	for i = 1,5 do
		local unit = "party"..i;
		if UnitExists(unit) then
			local partyMember = {}
			partyMember.guid = string.sub(UnitGUID(unit),8);
			partyMember.name = UnitFullName(unit);
			partyMember.specID = GetInspectSpecialization(unit) or 0;
			partyMember.raceID = select(3,UnitRace(unit));
			partyMember.classID = select(3,UnitClass(unit));
			partyMember.guildName = GetGuildInfo(unit);
			partyMember.role = GetSpecializationRoleByID(partyMember.specID);
			
			tinsert(partyMembers, partyMember)
		end
	end
	local runInfo = {
		["dungeonId"]= activeMythicPlusMap,
		["level"]=level,
		["completionTime"]=completionTime/1000.0,
		["deathCount"]=deathCount,
		["deathTimePenalty"]=deathTimePenalty,
		["enemyDamagerModifier"]=enemyDamagerModifier,
		["enemyHealthModifier"]=enemyHealthModifier,
		["onTime"]=onTime,
		["keystoneUpgradeLevels"]=keystoneUpgradeLevels,
		["seasonId"]=C_MythicPlus.GetCurrentSeason(),
		["weekIndex"]=KCLib:GetWeeklyIndex(),
		["timestamp"]=time(),
		["partyMembers"]=partyMembers
	}
	tinsert(KCLib:GetPlayerKeystone().runHistory, runInfo)
	
	C_MythicPlus.RequestRewards();
	C_MythicPlus.RequestMapInfo();
end

-- A frame for managing events.
local eventHandler = CreateFrame("Frame", nil, WorldFrame);

-- Trigger keystone updates after the delay has expired.
eventHandler:SetScript("OnUpdate", function()  
	if C_ChallengeMode.GetActiveChallengeMapID() then
		activeMythicPlusMap = C_ChallengeMode.GetActiveChallengeMapID();
		deathCount, deathTimePenalty = C_ChallengeMode.GetDeathCount()
	end
	if delayedRun and delayedRun > 0 and delayedRun < time() then
		delayedRun = 0;
		BroadcastPlayerKeys()
	end
end)

-- Pass events onto the correct function in 'EventCallables'.
eventHandler:SetScript("OnEvent", function (self, event, ...)
	EventCallables[event](...);
end)

-- Register all events in the 'EventCallables' table.
for event in pairs(EventCallables) do
	eventHandler:RegisterEvent(event);
end